! See copyright notice in the COPYRIGHT file.
! ****************************************************************************** !
!> This module contains data types, function and routines for turbulence
!! model. Especially the routine to compute turbulent eddy viscosity for
!! different turbulence model
!!
!! author: Kannan Masilamani
?? include 'header/lbm_macros.inc'
module mus_turbulence_module
  ! include treelm modules 
  use env_module,          only: rk, long_k, labelLen
  use tem_tools_module,    only: upper_to_lower, tem_horizontalSpacer
  use tem_aux_module,      only: tem_abort
  use tem_logging_module,  only: logUnit
  use tem_debug_module,    only: dbgUnit
  use tem_comm_module,     only: tem_communication_type, tem_commpattern_type, &
    &                            tem_comm_init
  use tem_varSys_module,   only: tem_varSys_type
  use tem_stencil_module,  only: tem_stencilHeader_type
  use tem_param_module,    only: qN00, q100, q0N0, q010, q00N, q001
  use tem_construction_module, only: tem_levelDesc_type
  
  ! include aotus modules
  use aotus_module,     only: aoterr_Fatal, aoterr_NonExistent, flu_State,     &
    &                         aoterr_WrongType, aot_top_get_val, aot_get_val
  use aot_table_module, only: aot_table_open, aot_table_close

  ! include musubi modules
  use mus_derVarPos_type_module, only: mus_derVarPos_type
  use mus_scheme_layout_module,  only: mus_scheme_layout_type

  implicit none
  private

  public :: mus_turbulence_type
  public :: mus_turbulence_config_type
  public :: mus_turbulence_data_type
  public :: mus_init_turbulenceData
  public :: mus_load_turbulence
  public :: mus_turb_calcVisc

  !> Contains large Eddy Turbulence (LES) model coefficients
  type les_coeff_type
    !> Smagorinsky constant.
    !! C_s = sqrt(C_k sqrt(C_k/C_e) ) = 0.16778594 = 0.17
    real(kind=rk) :: C_s
    !> Model constant for WALE (Wall-Adapting Local Eddy-Viscosity)
    !! default: 0.5
    real(kind=rk) :: C_w
    !> Model constant for Vreman model
    !! In literature: C_v = sqrt(2.5) C_s = 0.27
    real(kind=rk) :: C_v
    !> Modal constant for turbulent kinetic energy dissipation rate 
    !! default: 1.048
    real(kind=rk) :: C_e
    !> Model constant for eddy-viscosity coefficient
    !! default: 0.094_rk
    !! https://caefn.com/openfoam/smagorinsky-sgs-model
    real(kind=rk) :: C_k
  end type les_coeff_type

  !> Contains turbulence information loaded from config file
  type mus_turbulence_config_type
    !> turbulence model type
    character(len=labelLen) :: model

    !> les model coefficients
    type(les_coeff_type) :: coeff

    !> To compute strain-rate from non-equilibrium PDF for Smagorinsky les model.
    !! If true then velocity and grad data are not required
    logical :: compSR_fromPDF = .false.

    !> Use Kolmogorov scale for interpolation turbulent viscosity for multilevel 
    logical :: useKolmogorovScale = .false.
  end type mus_turbulence_config_type

  !> Contains information required to compute gradient like position
  !! of six direct neigbors in the state array and coeff for the gradient
  !! operation
  type mus_gradData_type
    !> Stores position of 6 direct face neighbors in the state array for
    !! each element
    !! Size: nSize, stencil%nDims, 2
    !! last index 2 refers to left and right side
    !! 1 is left/negative side and 2 is right/positive side
    integer, allocatable :: neighPos(:,:,:)

    !> coeff to calculate 1st order derivative
    !! use forward difference for element with boundary (coeff = 1.0)
    !! and central difference for inner fluid element (coeff=0.5)
    !! size: nSize, stencil%nDims
    real(kind=rk), allocatable :: FDcoeff(:,:)
  end type mus_gradData_type

  !> Contains velocity and gradient data to compute eddy viscosity
  type mus_turbulence_data_type
    !> Contains direct neighbor position in state and coeff to compute gradient
    type(mus_gradData_type) :: gradData

    !> Communication buffers to communicate visoscity field
    !! Local Fluids required by remote processes
    type( tem_communication_type ) :: sendBuffer
    !> My halos which are fluids on remote processes
    type( tem_communication_type ) :: recvBuffer
    !> Local ghostFromCoarser required by remote processes
    type( tem_communication_type ) :: sendBufferFromCoarser
    !> Local ghostFromFiner required by remote processes
    type( tem_communication_type ) :: sendBufferFromFiner
    !> My halos which are ghostFromCoarser on remote processes
    type( tem_communication_type ) :: recvBufferFromCoarser
    !> My halos which are ghostFromFiner on remote processes
    type( tem_communication_type ) :: recvBufferFromFiner

    !> Normalized turbulence viscosity 
    !! i.e. viscosity scaled to current level i.e. visc/dtL
    !! Size: nSize (nFluids+nGhosts+nHalos)
    !! Used gradData to compute viscosity for nFluids and nGhosts.
    !! This viscosity is interpolated and scaled for setting nonEq term 
    !! interpolation routines. The source element of interpolation might be
    !! halo so they are communicated.
    !! 
    !! Simple scaling assumping norm of strainrate tensor |S| in different 
    !! level is small: 
    !! Schneider, A. (2015). A Consistent Large Eddy Approach for 
    !! Lattice Boltzmann Methods and its Application to Complex Flows. 
    !! Technical University Kaiserslautern.
    !! v_c = 4 v_f. Scaled visc: v^s = v/dt.
    !! => v^s_c dtL_c = 4 v^s_f dtL_f => v^s_c = 2 v^s_f
    !! 
    !! Kolmogorov scaling:
    !! Touil, H., Ricot, D., & Lévêque, E. (2014). Direct and large-eddy 
    !! simulation of turbulent flows on composite multi-resolution grids by 
    !! the lattice Boltzmann method. Journal of Computational Physics, 256, 
    !! 220–233.
    !! v^s_c = 2^(1/3) v^s_f
    real(kind=rk), allocatable :: visc(:)
  end type mus_turbulence_data_type  


  !> Contains function pointers to obtain normalized turbulence viscosity.
  !! Viscosity is normalized to current level i.e. v_s = v/dt
  type mus_turbulence_visc_func_type
    !> this procedure compute eddy viscosity from velocity field depending
    !! turbulence and lbm (compressible/incompressible) models
    procedure(proc_calc_turb_visc_fromGradU), pointer, nopass &
      & :: calcFromGradU => null()
    
    !> this procedure compute eddy viscosity from preCollision PDF.
    !! It is used for Smagorinsky model which depends only on strain rate
    !! that can be calculated using local nonEquilibrium.
    !! Is assigned when compSR_fromPDF is .true.
    procedure(proc_calc_turb_visc_fromPreColPDF), pointer, nopass &
      & :: calcFromPreColPDF => null()
  end type mus_turbulence_visc_func_type

  !> Contains information required to compute eddy viscosity
  type mus_turbulence_type
    !> is true if turbulence table is defined
    logical :: active 
    !> exchange velocity if turbulence is active and viscosity
    !! is computed from velocity gradient
    logical :: exchangeVel
    !> information loaded from config file
    type(mus_turbulence_config_type) :: config
    !> contains level-wise turbulence data to compute eddy viscosity
    !! size: minlevel:maxLevel
    type(mus_turbulence_data_type), allocatable :: dataOnLvl(:)

    !> contains turbulence viscosity function pointers
    type(mus_turbulence_visc_func_type) :: viscFunc

    !> Factor to scale normalized turbulent viscosity from coarse to fine 
    !! depending on whether useKolmogorovScale true or false
    !! if useKolmogorovScale fac_c2f = 1/2^(1/3) else fac_c2f = 1/2
    !! How to use: v^s_f = fac_c2f v^s_c
    real(kind=rk) :: fac_c2f

    !> Factor to scale normalized turbulent viscosity from fine to coarse 
    !! depending on whether useKolmogorovScale true or false
    !! if useKolmogorovScale fac_f2c = 2^(1/3) else fac_f2c = 2
    !! How to use: v^s_c = fac_f2c v^s_f
    real(kind=rk) :: fac_f2c

  end type mus_turbulence_type

  !> interface to calculate subgrid scale turbulent eddy viscosity
  abstract interface
    !> This function computes turbulent viscosity from gradient U 
    pure function proc_calc_turb_visc_fromGradU(turbConfig, gradU, dxL, dtL) &
      & result(res)
      import :: rk, mus_turbulence_config_type

      !> turbulence config contains oefficients
      type(mus_turbulence_config_type), intent(in) :: turbConfig

      !> gradient of velocity
      !! Size: nDims, nDims
      real(kind=rk), intent(in) :: gradU(:,:)
 
      !> turbulence coefficients 
      !> current level lattice element size
      real(kind=rk), intent(in) :: dxL

      !> current level lattice time step size
      real(kind=rk), intent(in) :: dtL

      !> output: eddy viscosity of an element
      real(kind=rk) :: res
    end function proc_calc_turb_visc_fromGradU 

    !> This function compute turbulent viscosity from pre-collision PDF
    pure function proc_calc_turb_visc_fromPreColPDF(turbConfig, f_preCol, &
      & layout, dxL, dtL, viscKine) result(res)
      import :: rk, mus_turbulence_config_type, mus_scheme_layout_type

      !> turbulence type is implicitly passed to access turbulence coefficients
      type(mus_turbulence_config_type), intent(in) :: turbConfig

      !> pre-collision PDF
      !! size: QQ*nFields
      real(kind=rk), intent(in) :: f_preCol(:)

      !> scheme layout
      type(mus_scheme_layout_type), intent(in) :: layout

      !> current level lattice element size
      real(kind=rk), intent(in) :: dxL

      !> current level lattice time step size
      real(kind=rk), intent(in) :: dtL

      !> Background kinematic viscosity divided by dtL
      real(kind=rk), intent(in) :: viscKine

      !> output: eddy viscosity of an element
      real(kind=rk) :: res
    end function proc_calc_turb_visc_fromPreColPDF
  end interface  

contains 

  ! ************************************************************************** !
  !> load turbulence table
  subroutine mus_load_turbulence(me, conf, parent)
    !--------------------------------------------------------------------------
    !> fluid type
    type( mus_turbulence_type ),intent(out) :: me
    !> lua state
    type( flu_state )              :: conf
    !> parent handle
    integer, intent(in), optional  :: parent
    !--------------------------------------------------------------------------
    integer :: turb_table, iError
    !--------------------------------------------------------------------------

    ! if fluid informations in scheme table parentHandle /= 0
    if( present(parent) ) then
      call aot_table_open( L       = conf,        &
        &                  parent  = parent,      &
        &                  thandle = turb_table,  &
        &                  key     = 'turbulence' )
    else
      call aot_table_open( L=conf, thandle = turb_table, key = 'turbulence' )
    end if

    if ( turb_table == 0 ) then
      write(logUnit(1),*)'No turbulence table defined'
      me%config%compSR_fromPDF = .false.
      me%active = .false.
      me%exchangeVel = .false.
      return
    endif

    write(logUnit(1),*) 'Loading turbulence informations'
    me%active = .true.
    ! default is set to false for other models, its optional only for
    ! Smagorinsky model
    me%config%compSR_fromPDF = .false.
    me%exchangeVel = .true.

    ! load turbulence model
    call aot_get_val(L       = conf,            &
      &              thandle = turb_table,      &
      &              key     = 'model',         &
      &              val     = me%config%model, &
      &              default = 'smagorinsky',   &
      &              ErrCode = iError           )

    if (btest(iError, aoterr_Fatal)) then
      write(logUnit(1),*)'FATAL Error occured, while retrieving model :'
      if (btest(iError, aoterr_NonExistent)) then
        write(logUnit(1),*)'Variable not existent!'
        write(logUnit(1),*) 'Using default model: smagorinsky'
      end if  

      if (btest(iError, aoterr_WrongType)) then
         write(logUnit(1),*)'Variable has wrong type!'
         write(logUnit(1),*)'STOPPING'
         call tem_abort()
      end if  
    end if

    write(logUnit(1),*) '  model: ', trim(me%config%model)
    select case(upper_to_lower(trim(me%config%model)))
    case('smagorinsky')
      call aot_get_val(L       = conf,                         &
        &              thandle = turb_table,                   &
        &              key     = 'compute_strainrate_fromPDF', &
        &              val     = me%config%compSR_fromPDF,     &
        &              default = .true.,                       &
        &              ErrCode = iError                        )

      write(logUnit(1),*) '  use preCol PDF to calc strainrate: ', &
        &                 me%config%compSR_fromPDF
      ! use Smagorinksy constant when computing eddy viscosity from PDF
      ! else use C_k and C_e
      if (me%config%compSR_fromPDF) then
        call aot_get_val(L       = conf,                &
          &              thandle = turb_table,          &
          &              key     = 'c_s',               &
          &              val     = me%config%coeff%C_s, &
          &              default = 0.17_rk,             &
          &              ErrCode = iError               )
        write(logUnit(1),*) '  C_s: ', me%config%coeff%C_s

      else
        call aot_get_val(L       = conf,                &
          &              thandle = turb_table,          &
          &              key     = 'c_e',               &
          &              val     = me%config%coeff%C_e, &
          &              default = 1.048_rk,            &
          &              ErrCode = iError               )
        write(logUnit(1),*) '  C_e: ', me%config%coeff%C_e

        call aot_get_val(L       = conf,                &
          &              thandle = turb_table,          &
          &              key     = 'c_k',               &
          &              val     = me%config%coeff%C_k, &
          &              default = 0.094_rk,            &
          &              ErrCode = iError               )
        write(logUnit(1),*) '  C_k: ', me%config%coeff%C_k
      end if
    case('wale')
      call aot_get_val(L       = conf,                &
        &              thandle = turb_table,          &
        &              key     = 'c_w',               &
        &              val     = me%config%coeff%C_w, &
        &              default = 0.5_rk,              &
        &              ErrCode = iError               )


      write(logUnit(1),*) '  C_w: ', me%config%coeff%C_w
    case('vreman')
      ! load model coefficients
      call aot_get_val(L       = conf,                &
        &              thandle = turb_table,          &
        &              key     = 'c_v',               &
        &              val     = me%config%coeff%C_v, &
        &              default = 0.5_rk,              &
        &              ErrCode = iError               )
  
      write(logUnit(1),*) '  C_v: ', me%config%coeff%C_v
    case default  
        call tem_abort('Error: Unknown turbulence model')
    end select

    ! to use Kolmogorov scale for multilevel
    call aot_get_val(L       = conf,                         &
      &              thandle = turb_table,                   &
      &              key     = 'use_kolmogorov_scale',       &
      &              val     = me%config%useKolmogorovScale, &
      &              default = .true.,                       &
      &              ErrCode = iError                        )


    ! set scaling factor to convert turbulent viscosity from coarse to fine
    ! and vice versa
    if(me%config%useKolmogorovScale) then
      me%fac_f2c = 2.0_rk**(1.0_rk/3.0_rk)
      me%fac_c2f = 1.0_rk/me%fac_f2c
    else
      me%fac_f2c = 2.0_rk
      me%fac_c2f = 1.0_rk/me%fac_f2c
    end if
   
    ! deactivate exchange vel if SR is computed from PDF 
    if (me%config%compSR_fromPDF) me%exchangeVel = .false.
    
    call aot_table_close( L=conf, thandle=turb_table )
    call tem_horizontalSpacer(fUnit = logUnit(1))

  end subroutine mus_load_turbulence
  ! ************************************************************************** !


  ! ************************************************************************** !
  !> This initialize turbulence data type which includes velocity array,
  !! gradient data and communication buffer
  subroutine mus_init_turbulenceData(me, turbConfig, total, neigh, levelDesc, &
    &                                pattern, stencil, nElems, nSolve,   &
    &                                nScalars)
    !---------------------------------------------------------------------------
    !> turbulence data
    type(mus_turbulence_data_type), intent(out) :: me
    !> tubulence configuration
    type(mus_turbulence_config_type), intent(in) :: turbConfig
    !> total list
    integer(kind=long_k), intent(in) :: total(:)
    !> neighbor connectivity array
    integer, intent(in) :: neigh(:)
    !> levelDesc to access communication buffers of state array
    type(tem_levelDesc_type), intent(in) :: levelDesc
    !> communication pattern
    type(tem_commPattern_type), intent(in)      :: pattern
    !> stencil header
    type(tem_stencilHeader_type), intent(in) :: stencil
    !> Number of elements in state array
    integer, intent(in) :: nElems
    !> Number of elements solved in compute kernel i.e. excluding halos
    integer, intent(in) :: nSolve
    !> number of scalars in state array
    integer, intent(in) :: nScalars
    !---------------------------------------------------------------------------
    integer :: nDims
    !---------------------------------------------------------------------------
    nDims = stencil%nDims
    ! allocate tubulence viscosity array 
    allocate(me%visc(nElems))

    ! allocate velocity, gradData, send and recv buffer only for turbulence
    ! viscosity calculated using velocity field
    if (.not. turbConfig%compSR_fromPDF) then
    
      ! allocate gradient data type
      call gradData_allocate(me%gradData)
      
      ! fill gradient data with face neighbor information for turbulence model
      call fillGradData(me%gradData)

      ! initialize communication buffers for viscosity
      call init_commBuffer(buffer_visc  = me%sendBuffer,   &
        &                  buffer_State = levelDesc%sendBuffer )
      call init_commBuffer(buffer_visc  = me%recvBuffer,   &
        &                  buffer_State = levelDesc%recvBuffer )

      call init_commBuffer(buffer_visc  = me%sendBufferFromCoarser,      &
        &                  buffer_State = levelDesc%SendBufferFromCoarser)
      call init_commBuffer(buffer_visc  = me%recvBufferFromCoarser,      &
        &                  buffer_State = levelDesc%recvBufferFromCoarser)

      call init_commBuffer(buffer_visc  = me%sendBufferFromFiner,      &
        &                  buffer_State = levelDesc%sendBufferFromFiner)
      call init_commBuffer(buffer_visc  = me%recvBufferFromFiner,      &
        &                  buffer_State = levelDesc%recvBufferFromFiner)
    end if

    contains

    ! ************************************************************************ !
    subroutine init_commBuffer(buffer_visc, buffer_state)
      !-------------------------------------------------------------------------
      !> communication buffer for velocity field
      type(tem_communication_type), intent(out) :: buffer_visc
      !> communication buffer of state array which is already initialized
      !! in tem_construction_module
      type(tem_communication_type), intent(in) :: buffer_state
      !-------------------------------------------------------------------------
      ! positions, in velocity vectors
      integer, allocatable :: pos(:)
      integer :: iProc
      !-------------------------------------------------------------------------
      ! copy information about target and source procs from pdf sendBuffer to
      ! velocity sendBuffer
      call tem_comm_init(buffer_visc, buffer_state%nProcs)
      buffer_visc%proc = buffer_state%proc
      buffer_visc%nElemsProc = buffer_state%nElemsProc
      do iProc = 1, buffer_visc%nProcs
        call pattern%initBuf_real( me    = buffer_visc%buf_real( iProc ),   &
          &                        pos   = buffer_state%elemPos(iProc)%val, &
          &                        nVals = buffer_state%nElemsProc(iProc)   )
      end do     

    end subroutine init_commBuffer
    ! ************************************************************************ !


    ! ************************************************************************ !
    subroutine gradData_allocate( gradData )
      !---------------------------------------------------------------------------
      type(mus_gradData_type), intent(out) :: gradData
      !---------------------------------------------------------------------------
      allocate(gradData%neighPos( nSolve, nDims, 2 ) )
      gradData%neighPos = -1 
      allocate(gradData%FDcoeff( nSolve, nDims ) )
      gradData%FDcoeff = -1.0_rk
    end subroutine gradData_allocate
    ! ************************************************************************ !


    ! ************************************************************************ !
    !> This routine fills gradient data type with face neighbor information
    !! from pdf neigh array for turbulence model
    subroutine fillGradData( gradData )
      ! --------------------------------------------------------------------------
      type(mus_gradData_type), intent(inout) :: gradData
      ! --------------------------------------------------------------------------
      integer :: iElem, iDir, iFace, iMeshDir
      integer :: nFaces
      real(kind=rk) :: coeffForGrad
      integer :: nghDir, nghPos, nghElem
      ! --------------------------------------------------------------------------
      write(logUnit(1),*) 'Filling gradient data ...'
!write(dbgUnit(1),*) 'Filling gradient data ...'
      nFaces = stencil%nDims*2
  
      elemLoop: do iElem = 1, nSolve
!write(dbgUnit(1),*) 'iElem: ', iElem, ' treeID: ', total(iElem)
  
        iFace = 0
  
        ! Loop over every link
        neighLoop: do iDir  = 1, stencil%QQN
!write(dbgUnit(1),*) 'iDir: ', iDir        
  
          ! coeff to calculate 1st order derivative
          ! use forward difference for element with boundary (coeff = 1.0)
          ! and central difference for inner fluid element (coeff=0.5)
          coeffForGrad = 0.5_rk
  
          nghDir = ?NgDir?(iDir, stencil) 
          ! Find neighor using neigh array because for boundary, neigh array
          ! points to current element so no need to check for existence of the
          ! neighbor element
          nghElem = ?NghElemIDX?(neigh, iDir, stencil, iElem, nScalars, nElems)
  
          ! if neighbor element is my element then my neighbor is boundary
          ! use Forward difference to approximate gradient
          if (nghElem == iElem) coeffForGrad = 1.0_rk
  
          ! map neighbor direction to treelm stencil to identify the face
          iMeshDir = stencil%map(nghDir)
          select case(iMeshDir)
          case (qN00) !-x
            iFace = iFace+1
            gradData%neighPos(iElem,1,1) = nghElem
            gradData%FDcoeff(iElem,1) = coeffForGrad
          case (q100) !+x
            iFace = iFace+1
            gradData%neighPos(iElem,1,2) = nghElem
            gradData%FDcoeff(iElem,1) = coeffForGrad
          case (q0N0) !-y
            iFace = iFace+1
            gradData%neighPos(iElem,2,1) = nghElem
            gradData%FDcoeff(iElem,2) = coeffForGrad
          case (q010) !+y
            iFace = iFace+1
            gradData%neighPos(iElem,2,2) = nghElem
            gradData%FDcoeff(iElem,2) = coeffForGrad
          case (q00N) !-z
            iFace = iFace+1
            gradData%neighPos(iElem,3,1) = nghElem
            gradData%FDcoeff(iElem,3) = coeffForGrad
          case (q001) !+z
            iFace = iFace+1
            gradData%neighPos(iElem,3,2) = nghElem
            gradData%FDcoeff(iElem,3) = coeffForGrad
          end select  
!write(dbgUnit(1),*) 'iFace: ', iFace, 'iMeshDir: ', iMeshDir, 'coeff:', coeffForGrad
!write(dbgUnit(1),*) 'neighIDX: ', nghElem, ' neighID: ', total(nghElem)
          ! exit iDir loop 
          if (iFace == nFaces) exit neighLoop
        end do neighLoop
      end do elemLoop
!      write(logUnit(1),*) 'Done with fill grad data.'
!      write(dbgUnit(1),*) 'Done with fill grad data.'
!      flush(dbgUnit(1))
!      call tem_abort('Verify fill gradData')
    end subroutine fillGradData
    !************************************************************************* !

  end subroutine mus_init_turbulenceData
  ! ************************************************************************** !


  ! ************************************************************************** !
  !> This routine compute turbulence viscosity and stores in turbulence data type
  subroutine mus_turb_calcVisc(turbData, turbConfig, turbViscFunc, state, &
    &                          neigh, auxField, velPos, nSize, nSolve,    &
    &                          nScalars, nAuxScalars, layout, dxL, dtL,   &
    &                          viscKine)
    ! --------------------------------------------------------------------------
    !> turbulence data type
    type(mus_turbulence_data_type), intent(inout) :: turbData
    !> turbulence configuration
    type(mus_turbulence_config_type), intent(in) :: turbConfig
    !> turbulence function
    type(mus_turbulence_visc_func_type), intent(in) :: turbviscFunc
    !> state array
    real(kind=rk), intent(in) :: state(:)
    !> neigh array to obtain precollision pdf
    integer, intent(in) :: neigh(:)
    !> Auxiliary field variable array
    real(kind=rk), intent(in) :: auxField(:)
    !> position of velocity components in auxField
    integer, intent(in) :: velPos(3)
    !> number of elements in state array
    integer, intent(in) :: nSize
    !> Number of element to solve in this level
    integer, intent(in) :: nSolve
    !> number of scalars in state array
    integer, intent(in) :: nScalars
    !> number of scalars in auxField array
    integer, intent(in) :: nAuxScalars
    !> scheme layout
    type(mus_scheme_layout_type), intent(in) :: layout
    !> lattice element size in current level
    real(kind=rk), intent(in) :: dxL
    !> current level lattice time step size
    real(kind=rk), intent(in) :: dtL
    !> Background kinematic viscosity divided by dtL
    real(kind=rk), intent(in) :: viscKine(:)
    ! --------------------------------------------------------------------------
    integer :: iElem, iDir
    integer :: QQ, nDims
    real(kind=rk) :: pdf(layout%fStencil%QQ)
    real(kind=rk) :: gradU(3,3)
    ! --------------------------------------------------------------------------
    QQ = layout%fStencil%QQ
    nDims = layout%fStencil%nDims
    ! calculate turbulent viscosity
    if (turbConfig%compSR_fromPDF) then
      do iElem = 1, nSolve
        ! Get pre-collisiton PDF
        do iDir = 1, QQ
          pdf(iDir) = state (                                   &
           & ?FETCH?(iDir, 1, iElem, QQ, nScalars, nSize, neigh))
        end do

        ! compute normalized turbulence viscosity from nonEquilibrium PDF
        ! nonEquilibrium PDF is computed from preCollision depending on
        ! scheme kind
        turbData%visc(iElem)                                               &
          & = turbViscFunc%calcFromPreColPDF( turbConfig = turbConfig,     &
          &                                   f_preCol   = pdf,            &
          &                                   layout     = layout,         &
          &                                   dxL        = dxL,            &
          &                                   dtL        = dtL,            &
          &                                   viscKine   = viscKine(iElem) )
    
      end do !iElem
    else
      do iElem = 1, nSolve
        ! compute velocity gradient from velocity field stored in turbData
        gradU = getGradU( iElem = iElem ) 

        ! compute normalized turbulence viscosity
        turbData%visc(iElem)                                       &
          & = turbViscFunc%calcFromGradU( turbConfig = turbConfig, &
          &                               gradU      = gradU,      &
          &                               dxL        = dxL,        &
          &                               dtL        = dtL         )
      end do !iElem
    end if

    contains

    ! ************************************************************************ !
    !> This function computes gradient of velocity from gradient and veleocity 
    !! data.
    !! Gradient is computed using central difference.
    !! if an element has an boundary then neighbor refers to current element
    !! then forward difference is used
    pure function getGradU( iElem ) result(gradU)
      ! ------------------------------------------------------------------------
      !> element index to compute velocity gradient
      integer, intent(in) :: iElem
      !> output: gradient of velocity
      real(kind=rk) :: gradU(3, 3)
      ! ------------------------------------------------------------------------
      integer :: iRow, iCol
      integer :: leftNgh, rightNgh
      real(kind=rk) :: FDcoeff, leftVel(3), rightVel(3)
      ! ------------------------------------------------------------------------
      gradU = 0.0_rk
      do iCol = 1, nDims
        leftNgh = turbData%gradData%neighPos(iElem, iCol, 1)
        rightNgh = turbData%gradData%neighPos(iElem, iCol ,2)
        ! left and right velocity for each direction
        leftvel(1) = auxField( (leftNgh-1)*nAuxScalars+velPos(1) )
        leftvel(2) = auxField( (leftNgh-1)*nAuxScalars+velPos(2) )
        leftvel(3) = auxField( (leftNgh-1)*nAuxScalars+velPos(3) )

        rightvel(1) = auxField( (rightNgh-1)*nAuxScalars+velPos(1) )
        rightvel(2) = auxField( (rightNgh-1)*nAuxScalars+velPos(2) )
        rightvel(3) = auxField( (rightNgh-1)*nAuxScalars+velPos(3) )

        ! finite difference coefficients
        FDcoeff  = turbData%gradData%FDcoeff(iElem, iCol)
        do iRow = 1, nDims
          gradU(iRow, iCol) = (rightVel(iRow)-leftVel(iRow))*FDcoeff
        end do
      end do  
    end function getGradU
    ! ************************************************************************ !

  end subroutine mus_turb_calcVisc
  ! ************************************************************************** !

end module mus_turbulence_module
